import c4d
from c4d import plugins, bitmaps
import os


# Res ID's
# ========================================
PLUGIN_ID = 1060824  # Plugin ID
ICON_ID = 1060825  # Arch-Window Icon ID
ICON_ID_STOP = 1060829
ICON_ID_GRID = 1061024
ICON_ID_DUALGRID = 1061025
ICON_ID_STAR = 1061026
ICON_ID_BALCONY = 1061027
ICON_ID_ROOF = 1061028

# Res ID's
PY_WINDOW_LIST = 10001
PY_WINDOW_ARCH = 1

PY_WINDOW_WIDTH = 10004
PY_WINDOW_HEIGHT = 10005

# Global Parameter
PY_OFFSET = 10014
PY_BANK = 10015
PY_BANK_SIZE = 10016
PY_BANK_INNERSIZE = 10017
PY_RESET = 10018
PY_BAR_WIDTH = 10019
# Arch Window main parameters//////////////////////////
PY_GROUP_1 = 11000
PY_WINDOW_SIZE = 11001
PY_ARCH_HEIGHT = 11002
PY_ARCH_SUBD = 11003
PY_ARCH_FIXER = 11004
PY_ARCH_STRENGTH = 11005
PY_ARCH_POINTED = 11006
PY_WINDOW_SEGMENTS = 11007
# Arch Main Segmen
PY_TILT_CHECK = 11100
PY_WINDOW_OPEN = 11101
PY_WINDOW_DIRECTION = 11102
# Bars Arch left Segmen
PY_1BARS_LEFT = 11200
PY_1BARS_LEFTDEACTIVATED = 0
PY_1BARS_LEFTMETHOD1 = 1
PY_1BARS_LEFTMETHOD2 = 2
PY_1BARS_LEFTMETHOD3 = 3
PY_1BARS_LEFTVERT = 11201
PY_1BARS_LEFTHOR = 11202
PY_1BARS_LEFTMETHOD2DISTANCE = 11203
# new dual Grid ids lef
PY_1BARS_LEFTDUALVERT = 11204
PY_1BARS_LEFTDUALHOR = 11205
PY_1BARS_LEFTSTARSTREAKS = 11206
PY_1BARS_LEFTSTARRADIUS = 11217
PY_1BARS_LEFTSTARSUBD = 11218

# ======================================

# Handle Distance Values Frame Basic
HANDLE_X = 7  # Distance in cm from profile outside in x-axis
HANDLE_Z = 8.6  # Distance in cm from profile outside in z-axis

# Window Frame Type1 Point Datas ====================================
# Basic Frame Point Datas
AXIS_RIGHT = 17  # index point of window frame
AXIS_LEFT = 38  # index point of window frame
AXIS_BOTTOM = 80
AXIS_TOP = 79
ARCH_GLASS_POS = 7  # index point of window frame
FIXED_GLASS_POS = 4
CUTTER_POS = 5
FIXED_CUTTER_POS = 2
# # # # # # #

# im frame profile
FIXED_PROFILE_WIDTH = 8
WINDOW_HANDLE_POINT = 14

SILL_HEIGHT = 2  # width in cm
PROFILE_WIDTH = 12  # width in cm of profile in x-direction

# Frame Type 1 End ========================================================

HANDLE_DICT = {
    1: {
        1: {
            0: (37, 31, 15),
            1: (10, 16, 15),
            2: (),
            3: (),
            4: ()
        },
        2: {
            0: (37, 31, 15),
            1: (10, 16, 15),
            2: (10, 16, 15),
            3: (10, 16, 15),
            4: ()
        },
        3: {
            0: (37, 31, 15),
            1: (10, 16, 15),
            2: (10, 16, 15),
            3: (),
            4: ()
        },
        4: {
            0: (10, 16, 15),
            1: (10, 16, 15),
            2: (),
            3: (),
            4: ()
        },
        5: {
            0: (10, 16, 15),
            1: (10, 16, 15),
            2: (),
            3: (),
            4: ()
        },
        6: {
            0: (37, 31),
            1: (10, 16),
            2: (),
            3: (),
            4: ()
        },
        7: {
            0: (37, 31),
            1: (10, 16),
            2: (),
            3: (),
            4: ()
        }
    }
}


class WindowHelper:
    def __init__(self):
        pass

    @staticmethod
    def create_tex_tags(op, name):
        tag = c4d.BaseTag(c4d.Ttexture)
        tag[c4d.TEXTURETAG_PROJECTION] = 6
        tag[c4d.ID_BASELIST_NAME] = name
        op.InsertTag(tag)

    @staticmethod
    def catch_tex_tag(op, name, goal):
        # new tag read from op
        frame_tag = None
        for tag in op.GetTags():
            if tag.GetName().lower() == name:
                frame_tag = tag
                break
        if frame_tag is not None:
            if goal:
                goal.InsertTag(frame_tag.GetClone())

    @staticmethod
    def load_profile():
        frame_path = os.path.join(path, "res", "w_f")
        window_path = os.path.join(path, "res", "w_w")
        lock_path = os.path.join(path, "res", "w_l")
        frame_profile = c4d.BaseObject(c4d.Ospline)
        window_profile = c4d.BaseObject(c4d.Ospline)
        lock_profile = c4d.BaseObject(c4d.Ospline)
        c4d.storage.ReadHyperFile(None, frame_profile, frame_path, 100)
        c4d.storage.ReadHyperFile(None, window_profile, window_path, 100)
        c4d.storage.ReadHyperFile(None, lock_profile, lock_path, 100)
        return frame_profile, window_profile, lock_profile

    @staticmethod
    def set_handle_pos(op, frame, handle, f_type=1, w_type=1, h_pos=0):

        point_data = HANDLE_DICT[f_type][w_type][h_pos]  # tuple mit (vertex 1, vertex 2)
        if h_pos != 4:
            frame = frame.GetDown().GetDown().GetDown()
        else:
            pass

        if f_type == 1:
            # Frame Type 1
            if w_type == 1:
                # Window Type Arch 1 Segment  / works also with 2 Segments

                pos_x = (frame.GetPoint(point_data[0]).x - frame.GetPoint(point_data[1]).x) / 2
                pos_x += frame.GetPoint(point_data[1]).x
                pos_z = frame.GetPoint(point_data[2]).z

                if h_pos == 0:
                    point = point_data[1]
                    pos_y = frame.GetPoint(point).y + ((frame.GetPoint(point + 21).y - frame.GetPoint(point).y) / 2)
                elif h_pos == 1:
                    point = point_data[0]
                    if not op[PY_WINDOW_SEGMENTS]:
                        pos_y = frame.GetPoint(point).y + (
                                (frame.GetPoint(point + (2 * 21)).y - frame.GetPoint(point).y) / 2)
                    else:
                        pos_y = frame.GetPoint(point).y + (
                                (frame.GetPoint(frame.GetPointCount() - 11).y - frame.GetPoint(point).y) / 2)

                pos = c4d.Vector(pos_x, pos_y, pos_z)
                handle.SetRelPos(pos)

    @staticmethod
    def load_handle():
        handle_path = os.path.join(path, "res", "w_h")
        handle = c4d.BaseObject(c4d.Opolygon)
        c4d.storage.ReadHyperFile(None, handle, handle_path, 100)

        return handle

    @staticmethod
    def load_sill():
        sill_path = os.path.join(path, "res", "b_al")
        sill = c4d.BaseObject(c4d.Ospline)
        c4d.storage.ReadHyperFile(None, sill, sill_path, 100)
        return sill

    @staticmethod
    def set_axis(op, pos):
        ax_pos = c4d.Vector()
        if pos == 0:  # left
            ax_pos = c4d.Vector(0)
        elif pos == 1:  # middle
            ax_pos = c4d.Vector(-(op[PY_WINDOW_WIDTH] / 2), 0, 0)
        elif pos == 2:  # right
            ax_pos = c4d.Vector(-op[PY_WINDOW_WIDTH], 0, 0)
        elif pos == 3:  # midpoint
            if op[PY_WINDOW_LIST] != 3:
                ax_pos = c4d.Vector(-op[PY_WINDOW_WIDTH], -op[PY_WINDOW_HEIGHT], 0) / 2
            else:
                ax_pos = c4d.Vector(-op[PY_WINDOW_WIDTH], -op[PY_WINDOW_WIDTH], 0) / 2

        return ax_pos

    @staticmethod
    def create_arch_spline(op, widtho, heighto, archo, subdo, multi=1, glass=False):
        height = heighto
        width = widtho
        if archo is not None:
            arch_height = archo
        else:
            arch_height = 20
        subd_points = subdo

        if op[PY_BANK]:
            if glass:
                sill_pos = 0
                height -= 2
            else:
                sill_pos = SILL_HEIGHT
        else:
            sill_pos = 0

        # automatic adjustment of height
        if arch_height > width / 2:
            arch_height = width / 2

        spline = c4d.BaseObject(c4d.Ospline)
        spline[c4d.SPLINEOBJECT_CLOSED] = True

        if arch_height == 0:
            subd_points = 0
        else:
            if subd_points == 0:
                subd_points = 1

        if subd_points == 0:
            points = 4
        else:
            points = 2 * subd_points + 5
        spline.ResizeObject(points)

        # if subd_points > 0:
        if arch_height != 0:
            radius = (4 * (arch_height ** 2) + width ** 2) / (8 * arch_height)
        else:
            radius = 0

        y_value = height - arch_height
        spline.SetPoint(0, c4d.Vector(0, sill_pos, 0))
        spline.SetPoint(1, c4d.Vector(width, sill_pos, 0))
        spline.SetPoint(2, c4d.Vector(width, y_value, 0))
        spline.SetPoint(points - 1, c4d.Vector(0, y_value, 0))

        # this is for the rounded arch
        if subd_points > 0:
            spline.SetPoint(2 + subd_points + 1, c4d.Vector(width / 2, height, 0))
            midpoint_pos = c4d.Vector(width / 2, height - radius, 0)
            rv_corner = spline.GetPoint(2) - midpoint_pos
            rv_right = spline.GetPoint(2 + subd_points + 1) - midpoint_pos
            first_angle = c4d.utils.GetAngle(rv_right, rv_corner)  # rv_corner.GetAngle(rv_right)

            angle_adder = first_angle / subd_points
            angle = c4d.utils.DegToRad(90) - angle_adder

            if height / arch_height > 10:
                factor = 3 * multi * 0.32 * subd_points  # innen

            elif height / arch_height > 6:
                factor = 7 * multi * 0.25 * subd_points

            else:
                factor = 7 * multi * 0.4 * subd_points

            for index in range(subd_points):
                sine, cose = c4d.utils.SinCos(angle)
                y = sine * radius
                x = cose * radius
                new_point = c4d.Vector(width / 2 + x, midpoint_pos.y + y, 0)
                spline.SetPoint(2 + subd_points - index, new_point)

                if not glass:
                    if index < 2 + subd_points - index:
                        if op[PY_ARCH_FIXER]:
                            angle_adder -= angle_adder / (60 / op[PY_ARCH_STRENGTH] * factor)

                angle -= angle_adder

            spline.SetPoint(2 + subd_points + 1, c4d.Vector(width / 2, height, 0))

            angle_adder = first_angle / subd_points
            angle = c4d.utils.DegToRad(90) + angle_adder

            if height / arch_height > 10:
                factor = 3 * multi * 0.32 * subd_points  # innen

            elif height / arch_height > 6:
                factor = 7 * multi * 0.25 * subd_points

            else:
                factor = 7 * multi * 0.4 * subd_points

            for index in range(subd_points):

                sine, cose = c4d.utils.SinCos(angle)
                y = sine * radius
                x = cose * radius
                new_point = c4d.Vector(midpoint_pos.x + x, midpoint_pos.y + y, 0)
                spline.SetPoint(2 + subd_points + 2 + index, new_point)
                if not glass:
                    if index < 2 + subd_points - index:
                        if op[PY_ARCH_FIXER]:
                            angle_adder -= angle_adder / (60 / op[PY_ARCH_STRENGTH] * factor)

                angle += angle_adder

        return spline

    def make_frames(self, node, f_spline, w_spline=None, pos=0, tilt=False, window_open=0):

        axis_position = AXIS_LEFT
        if pos != 4:
            # if frame is not fixed
            frame_spline = f_spline
            window_spline = w_spline
            if pos == 1:
                axis_position = AXIS_LEFT  # the index of the frame where to position the axis
            elif pos == 0:
                axis_position = AXIS_RIGHT
            elif pos == 2:
                axis_position = AXIS_LEFT
            elif pos == 3:
                axis_position = AXIS_BOTTOM

            frame_sweep = c4d.BaseObject(c4d.Osweep)
            frame_sweep[c4d.SWEEPOBJECT_ROTATIONFROM] = c4d.utils.DegToRad(-180)
            frame_sweep[c4d.SWEEPOBJECT_ROTATIONTO] = c4d.utils.DegToRad(180)

            frame_sweep.SetName("Frame")

            frame_spline.InsertUnder(frame_sweep)

            window_sweep = c4d.BaseObject(c4d.Osweep)
            window_sweep[c4d.SWEEPOBJECT_ROTATIONFROM] = c4d.utils.DegToRad(-180)
            window_sweep[c4d.SWEEPOBJECT_ROTATIONTO] = c4d.utils.DegToRad(180)
            window_sweep.SetName("Window")
            if window_spline:
                window_spline.InsertUnder(window_sweep)

            frame_profile = self.profile[0].GetClone()

            window_profile = self.profile[1].GetClone()

            frame_profile.InsertUnder(frame_sweep)
            window_profile.InsertUnder(window_sweep)

            frame_sweep: c4d.BaseObject = \
                c4d.utils.SendModelingCommand(command=c4d.MCOMMAND_CURRENTSTATETOOBJECT, list=[frame_sweep.GetClone()],
                                              doc=c4d.documents.BaseDocument())[0]
            window_sweep: c4d.BaseObject = \
                c4d.utils.SendModelingCommand(command=c4d.MCOMMAND_CURRENTSTATETOOBJECT, list=[window_sweep.GetClone()],
                                              doc=c4d.documents.BaseDocument())[0]

            frame_sweep.SetPhong(on=True, anglelimit=True, angle=c4d.utils.DegToRad(40))

            window_sweep.SetPhong(on=True, anglelimit=True, angle=c4d.utils.DegToRad(40))

            # new tag read from op
            WindowHelper().catch_tex_tag(node, "frame", frame_sweep)
            WindowHelper().catch_tex_tag(node, "frame", window_sweep)

            position = window_sweep.GetPoint(axis_position)
            window_axis = c4d.BaseObject(c4d.Onull)
            window_axis.SetName("Axis")
            container = c4d.BaseObject(c4d.Onull)
            frame_sweep.InsertUnder(container)
            window_axis.InsertUnder(container.GetDown())
            window_axis.SetAbsPos(position)
            window_sweep.InsertUnder(window_axis)
            window_sweep.SetRelPos(container.GetAbsPos() - position)

            if tilt:
                if pos == 1:
                    if node[PY_WINDOW_OPEN] is not None:
                        window_axis.SetRelRot(c4d.Vector(0, c4d.utils.DegToRad(-5), 0))
                elif pos == 0:
                    if node[PY_WINDOW_OPEN] is not None:
                        window_axis.SetRelRot(c4d.Vector(0, c4d.utils.DegToRad(-5), 0))
                elif pos == 2:
                    if node[PY_WINDOW_OPEN] is not None:
                        window_axis.SetRelRot(c4d.Vector(0, c4d.utils.DegToRad(-5), 0))
                elif pos == 3:
                    if node[PY_WINDOW_OPEN] is not None:
                        window_axis.SetRelRot(c4d.Vector(0, c4d.utils.DegToRad(5), 0))
            else:
                if pos == 1:
                    window_axis.SetRelRot(c4d.Vector(-window_open, 0, 0))
                elif pos == 0:
                    window_axis.SetRelRot(c4d.Vector(window_open, 0, 0))
                elif pos == 2:
                    window_axis.SetRelRot(c4d.Vector(0, -window_open, 0))
                elif pos == 3:
                    window_axis.SetRelRot(c4d.Vector(0, window_open, 0))

        else:
            # if frame is fixed
            frame_spline = f_spline

            frame_sweep = c4d.BaseObject(c4d.Osweep)
            frame_sweep[c4d.SWEEPOBJECT_ROTATIONFROM] = c4d.utils.DegToRad(-180)
            frame_sweep[c4d.SWEEPOBJECT_ROTATIONTO] = c4d.utils.DegToRad(180)

            frame_sweep.SetName("Frame")

            frame_spline.InsertUnder(frame_sweep)

            frame_profile = self.profile[2].GetClone()

            frame_profile.InsertUnder(frame_sweep)

            frame_sweep: c4d.BaseObject = \
                c4d.utils.SendModelingCommand(command=c4d.MCOMMAND_CURRENTSTATETOOBJECT, list=[frame_sweep.GetClone()],
                                              doc=c4d.documents.BaseDocument())[0]

            frame_sweep.SetPhong(on=True, anglelimit=True, angle=c4d.utils.DegToRad(40))

            # new tag read from op
            WindowHelper().catch_tex_tag(node, "frame", frame_sweep)

            container = c4d.BaseObject(c4d.Onull)
            frame_sweep.InsertUnder(container)

        return container

    @staticmethod
    def create_sill(op):
        spline: c4d.SplineObject = c4d.BaseObject(c4d.Ospline)
        spline[c4d.SPLINEOBJECT_CLOSED] = True
        spline.ResizeObject(6)
        spline.SetPoint(0, c4d.Vector(0, 0, 0))
        spline.SetPoint(1, c4d.Vector(0, 0, -10 - op[PY_BANK_SIZE]))
        spline.SetPoint(2, c4d.Vector(op[PY_WINDOW_WIDTH], 0, -10 - op[PY_BANK_SIZE]))
        spline.SetPoint(3, c4d.Vector(op[PY_WINDOW_WIDTH], 0, 0))
        spline.SetPoint(4, c4d.Vector(op[PY_WINDOW_WIDTH], 0, 15 + op[PY_BANK_INNERSIZE]))
        spline.SetPoint(5, c4d.Vector(0, 0, 15 + op[PY_BANK_INNERSIZE]))
        extrude = c4d.BaseObject(c4d.Oextrude)
        extrude.SetName("Sill")
        extrude[c4d.EXTRUDEOBJECT_DIRECTION] = SILL_HEIGHT
        extrude[c4d.EXTRUDEOBJECT_EXTRUSIONOFFSET] = SILL_HEIGHT
        spline.InsertUnder(extrude)
        extrude = c4d.utils.SendModelingCommand(command=c4d.MCOMMAND_CURRENTSTATETOOBJECT, list=[extrude],
                                                doc=op.GetDocument())[0]

        for tag in extrude.GetTags():
            if tag.IsInstanceOf(c4d.Tpolygonselection) or tag.IsInstanceOf(c4d.Tedgeselection):
                tag.Remove()

        WindowHelper().catch_tex_tag(op, "sill", extrude)
        return extrude

    def make_sill(self, op):
        extrude: c4d.BaseObject = c4d.BaseObject(c4d.Oextrude)
        extrude[c4d.EXTRUDEOBJECT_EXTRUSIONOFFSET] = - op[PY_WINDOW_WIDTH]

        sill = self.sill.GetClone()
        vector_z = op[PY_BANK_SIZE]
        if vector_z < 0:
            vector_z *= -1
        for point in range(sill.GetPointCount()):
            if point == 6 or point == 7:
                continue

            sill.SetPoint(point, sill.GetPoint(point) - c4d.Vector(0, 0, vector_z))
            vector_z = op[PY_BANK_SIZE]

        sill.InsertUnder(extrude)
        extrude = c4d.utils.SendModelingCommand(command=c4d.MCOMMAND_CURRENTSTATETOOBJECT,
                                                doc=op.GetDocument(),

                                                list=[extrude])[0]
        for tag in extrude.GetTags():
            if tag.IsInstanceOf(c4d.Tpolygonselection) or tag.IsInstanceOf(c4d.Tedgeselection):
                tag.Remove()

        extrude.SetName("Sill")
        extrude.SetPhong(on=True, anglelimit=True, angle=c4d.utils.DegToRad(60))
        return extrude

    @staticmethod
    def create_ring(op, radius=25, full_spiral=True, sub_d=8):

        tube = c4d.BaseObject(c4d.Otube)
        tube[c4d.PRIM_AXIS] = 5
        tube[c4d.PRIM_TUBE_ORAD] = radius
        if not full_spiral:
            tube[c4d.PRIM_SLICE] = True
        tube[c4d.PRIM_TUBE_IRAD] = radius - op[PY_BAR_WIDTH]
        tube[c4d.PRIM_TUBE_SEG] = 4 * sub_d
        tube[c4d.PRIM_TUBE_HEIGHT] = 2 - 0.01
        tube[c4d.PRIM_TUBE_HSUB] = 1
        # pos = tube.GetRelPos() - c4d.Vector(0, 0, (2 - 0.01) / 2 + 0.01)
        # tube.SetRelPos(pos)

        tube: c4d.PolygonObject = c4d.utils.SendModelingCommand(command=c4d.MCOMMAND_CURRENTSTATETOOBJECT, list=[tube],
                                                                doc=op.GetDocument())[0]
        for tag in tube.GetTags():
            if tag.IsInstanceOf(c4d.Tpolygonselection) or tag.IsInstanceOf(c4d.Tedgeselection):
                tag.Remove()

        for index in range(len(tube.GetAllPoints())):
            tube.SetPoint(index, tube.GetPoint(index) - c4d.Vector(0, 0, (2 - 0.01) / 2 + 0.01))
        return tube

    @staticmethod
    def create_bar(op, length, width, angle=0):
        """ Creates a bar with the length"""

        spline = c4d.SplineObject(4, c4d.SPLINETYPE_LINEAR)
        spline[c4d.SPLINEOBJECT_CLOSED] = True

        if angle == 0:
            bar = c4d.BaseObject(c4d.Oextrude)
            spline.SetPoint(0, c4d.Vector(-width / 2, 0, -.01))
            spline.SetPoint(1, c4d.Vector(width / 2, 0, -.01))
            spline.SetPoint(2, c4d.Vector(width / 2, 0, -2))
            spline.SetPoint(3, c4d.Vector(-width / 2, 0, -2))
            bar[c4d.EXTRUDEOBJECT_DIRECTION] = 2
            bar[c4d.EXTRUDEOBJECT_EXTRUSIONOFFSET] = length
            spline.InsertUnder(bar)

        elif angle == 90:
            bar = c4d.BaseObject(c4d.Oloft)
            bar[c4d.LOFTOBJECT_SUBX] = 4
            bar[c4d.LOFTOBJECT_SUBY] = 2
            bar[c4d.LOFTOBJECT_ADAPTIVEY] = False
            spline.SetPoint(0, c4d.Vector(0, width / 2, -.01))
            spline.SetPoint(1, c4d.Vector(0, -width / 2, -.01))
            spline.SetPoint(2, c4d.Vector(0, -width / 2, -2))
            spline.SetPoint(3, c4d.Vector(0, width / 2, -2))
            spline2 = spline.GetClone()
            spline.InsertUnder(bar)
            spline2.InsertUnder(bar)
            spline.SetRelPos(c4d.Vector(-length / 2, 0, 0))
            spline2.SetRelPos(c4d.Vector(length / 2, 0, 0))

        bar.SetName("Bar")

        bar: c4d.PolygonObject = c4d.utils.SendModelingCommand(command=c4d.MCOMMAND_CURRENTSTATETOOBJECT, list=[bar],
                                                               doc=op.GetDocument())[0]
        for tag in bar.GetTags():
            if tag.IsInstanceOf(c4d.Tpolygonselection) or tag.IsInstanceOf(c4d.Tedgeselection):
                tag.Remove()

        return bar

    @staticmethod
    def bars_maker(op, width, height, method=0, fixed=False, bank=False, vert=0, hor=0, height2=0, count=2,
                   full_spiral=True):
        container = c4d.BaseObject(c4d.Onull)
        if fixed != 4:
            pro_width = PROFILE_WIDTH
            z = 4
        else:
            pro_width = FIXED_PROFILE_WIDTH
            z = 2.7
        bar_width = op[PY_BAR_WIDTH]
        if method == 0:
            return None
        elif method == 1:
            # if not fixed:
            offset_vert = (width - 2 * pro_width) / (vert + 1)

            if not bank:
                offset_hor = (height - 2 * pro_width) / (hor + 1)
            else:
                offset_hor = (height + SILL_HEIGHT - 2 * pro_width) / (hor + 1)

            for index in range(vert):
                pos_x = pro_width + (index + 1) * offset_vert

                length = height
                bar = WindowHelper().create_bar(op, length=length, width=bar_width)
                bar.InsertUnder(container)

                bar.SetRelPos(c4d.Vector(pos_x, 0, z))

            # horizontal length calc inside arch

            for index in range(hor):
                pos_y = pro_width + ((index + 1) * offset_hor)
                length = width
                bar = WindowHelper().create_bar(op, length=length, width=bar_width, angle=90)
                bar.InsertUnder(container)
                bar.SetRelPos(c4d.Vector(width / 2, pos_y, z))

            return container

        elif method == 2:
            if bank:
                y = SILL_HEIGHT
            else:
                y = 0
            offset_vert = (width - 2 * pro_width) / (vert + 1)
            offset_vert2 = (width - 2 * pro_width) / (hor + 1)

            bar_middle = WindowHelper().create_bar(op, length=width, width=bar_width, angle=90)
            bar_middle.InsertUnder(container)
            bar_middle.SetRelPos(c4d.Vector(width / 2, height2 + y, z))

            for index in range(vert):
                pos_x_bottom = pro_width + (index + 1) * offset_vert
                length = height2 - bar_width / 2
                bar = WindowHelper().create_bar(op, length=length + y, width=bar_width)
                bar.InsertUnder(container)
                bar.SetRelPos(c4d.Vector(pos_x_bottom, 0, z))

            for index in range(hor):
                pos_x_top = pro_width + (index + 1) * offset_vert2
                length = height - height2 - bar_width / 2
                bar = WindowHelper().create_bar(op, length=length, width=bar_width)
                bar.InsertUnder(container)
                bar.SetRelPos(c4d.Vector(pos_x_top, height2 + y + bar_width / 2, z))

            return container

        elif method == 3:
            radius = height2
            if hor == 0:
                sub_d = 1
            else:
                sub_d = hor * 3

            if full_spiral:
                y_pos = height / 2
            else:
                if fixed != 4:
                    y_pos = PROFILE_WIDTH - 1
                else:
                    y_pos = FIXED_PROFILE_WIDTH

            ring = WindowHelper().create_ring(op, radius=radius, full_spiral=full_spiral, sub_d=sub_d)

            ring.InsertUnder(container)
            ring.SetRelPos(c4d.Vector(width / 2, y_pos, z))

            bar_boole = c4d.BaseObject(c4d.Oboole)
            bar_boole[c4d.BOOLEOBJECT_SINGLE_OBJECT] = True
            bar_boole[c4d.BOOLEOBJECT_HIDE_NEW_EDGES] = True

            bar_cutter = c4d.BaseObject(c4d.Ocylinder)
            bar_cutter[c4d.PRIM_AXIS] = 5
            if not full_spiral:
                bar_cutter[c4d.PRIM_SLICE] = True
            bar_cutter[c4d.PRIM_CYLINDER_HSUB] = 1
            bar_cutter[c4d.PRIM_CYLINDER_SEG] = 4 * sub_d
            bar_cutter[c4d.PRIM_CYLINDER_RADIUS] = radius

            bar_container = c4d.BaseObject(c4d.Onull)
            if width > height:
                length = width
            else:
                length = height
            if full_spiral:

                if count <= 2:
                    # create the spiral bars

                    bar = WindowHelper().create_bar(op, length=length, width=bar_width, angle=0)
                    bar_2 = bar.GetClone()
                    bar.InsertUnder(bar_container)
                    bar_2.InsertUnder(bar_container)
                    if count <= 1:
                        bar.SetRelRot(c4d.Vector(0, 0, c4d.utils.DegToRad(90)))
                        bar_2.SetRelRot(c4d.Vector(0, 0, c4d.utils.DegToRad(-90)))
                    else:
                        bar.SetRelRot(c4d.Vector(0, 0, c4d.utils.DegToRad(0)))
                        bar_2.SetRelRot(c4d.Vector(0, 0, c4d.utils.DegToRad(180)))

                else:
                    angle_adder = 360.0 / count
                    angle = 90 - angle_adder
                    for index in range(count):
                        bar = WindowHelper().create_bar(op, length=length, width=bar_width, angle=0)
                        bar.InsertUnder(bar_container)

                        bar.SetRelRot(c4d.Vector(0, 0, c4d.utils.DegToRad(angle + abs(angle_adder - 90))))
                        angle += angle_adder

                    if count > 4:
                        bar_container.SetRelRot(c4d.Vector(0, 0, c4d.utils.DegToRad(180)))
            else:

                if count < 1:
                    count = 1
                angle_adder = 180.0 / (count + 1)
                angle = 0 + angle_adder
                for index in range(count):
                    bar = WindowHelper().create_bar(op, length=length, width=bar_width, angle=0)
                    bar.InsertUnder(bar_container)

                    bar.SetRelRot(c4d.Vector(0, 0, c4d.utils.DegToRad(angle - 90)))
                    angle += angle_adder

            bar_cutter.InsertUnder(bar_boole)
            bar_container.InsertUnder(bar_boole)
            bar_boole.InsertUnder(container)
            bar_boole.SetRelPos(c4d.Vector(width / 2, y_pos, z))

            return container

    @staticmethod
    def bar_cutter(op, obj, bars, fixed=False):
        normal = False
        direction = 5
        if fixed != 4:
            frame: c4d.BaseObject = obj.GetDown().GetDown().GetDown()
            numbers = frame.GetPointCount() // 21
            points_list = []
            start = CUTTER_POS
            for point in range(numbers):
                points_list.append(frame.GetPoint(start))
                start += 21
        else:
            # print(obj.GetDown())
            direction = 5
            normal = False
            frame: c4d.BaseObject = obj.GetDown()
            numbers = frame.GetPointCount() // 12
            points_list = []
            start = FIXED_CUTTER_POS
            for point in range(numbers):
                points_list.append(frame.GetPoint(start))
                start += 12

        glass_spline = c4d.SplineObject(len(points_list), type=c4d.SPLINETYPE_LINEAR)
        glass_spline.SetAllPoints(points_list)
        glass_spline[c4d.SPLINEOBJECT_CLOSED] = True
        extrude = c4d.BaseObject(c4d.Oextrude)
        extrude[c4d.EXTRUDEOBJECT_FLIPNORMALS] = normal
        extrude[c4d.EXTRUDEOBJECT_EXTRUSIONOFFSET] = direction
        extrude.SetName("Glass")
        glass_spline.InsertUnder(extrude)

        # doc = c4d.documents.BaseDocument()
        # doc.InsertObject(extrude)
        # extrude = doc.GetFirstObject()

        extrude: c4d.PolygonObject = c4d.utils.SendModelingCommand(command=c4d.MCOMMAND_CURRENTSTATETOOBJECT,
                                                                   list=[extrude.GetClone()],
                                                                   doc=op.GetDocument())[0]

        for tag in extrude.GetTags():
            if tag.IsInstanceOf(c4d.Tpolygonselection) or tag.IsInstanceOf(c4d.Tedgeselection):
                tag.Remove()

        boolen = c4d.BaseObject(c4d.Oboole)
        boolen[c4d.BOOLEOBJECT_TYPE] = 2
        boolen[c4d.BOOLEOBJECT_SINGLE_OBJECT] = True
        boolen[c4d.BOOLEOBJECT_BREAK_CUT_EDGES] = True

        if not bars.GetDown():
            bars = c4d.BaseObject(c4d.Onull)
            bars.SetName("Bars")
            bars.InsertUnder(frame)
        else:
            bars.InsertUnder(boolen)
            extrude.InsertUnder(boolen)

            # doc = c4d.documents.BaseDocument()
            # doc.InsertObject(boolen)
            # bars = doc.GetFirstObject().GetClone()

            # bars: c4d.BaseObject = c4d.utils.SendModelingCommand(command=c4d.MCOMMAND_CURRENTSTATETOOBJECT,
            #                                                      list=[boolen],
            #                                                      doc=op.GetDocument())[0]
            #
            # if not bars.IsInstanceOf(c4d.Oboole):
            #     bars.SetPhong(True, anglelimit=True, angle=c4d.utils.DegToRad(40))
            #
            #     bars.SetName("Bars")
            #     for tag in bars.GetTags():
            #         if tag.IsInstanceOf(c4d.Tedgeselection):
            #             tag.Remove()
            #     WindowHelper().catch_tex_tag(op, "bars", bars)
            #
            #     bars.InsertUnder(frame)

            WindowHelper().catch_tex_tag(op, "bars", boolen)
            boolen.InsertUnder(frame)


    def window_arch(self, op):
        container = c4d.BaseObject(c4d.Onull)
        # self.create_arch_window(op)

        spline = self.create_arch_spline(op, op[PY_WINDOW_WIDTH], op[PY_WINDOW_HEIGHT],
                                         op[PY_ARCH_HEIGHT],
                                         op[PY_ARCH_SUBD], multi=1.5)

        w_spline = self.create_arch_spline(op, op[PY_WINDOW_WIDTH], op[PY_WINDOW_HEIGHT],
                                           op[PY_ARCH_HEIGHT],
                                           op[PY_ARCH_SUBD], multi=.95)
        frames = self.make_frames(op, spline, w_spline, op[PY_WINDOW_DIRECTION], tilt=op[PY_TILT_CHECK],
                                  window_open=op[PY_WINDOW_OPEN])

        # glass creation
        self.preshious_glass(op, frames)

        # handle creation
        handle = self.handle.GetClone()
        WindowHelper().catch_tex_tag(op, "handle", handle)  # Mats for handle
        handle.InsertUnder(frames.GetDown().GetDown().GetDown())

        # test new handel function
        self.set_handle_pos(op, frames, handle, w_type=1, h_pos=op[PY_WINDOW_DIRECTION])

        # new bars IDs =========================================
        if op[PY_1BARS_LEFT] > 0:
            if op[PY_1BARS_LEFT] == 1:
                # Grid
                bars_container = self.bars_maker(op, width=op[PY_WINDOW_WIDTH],
                                                 height=op[PY_WINDOW_HEIGHT],
                                                 method=op[PY_1BARS_LEFT],
                                                 bank=op[PY_BANK],
                                                 vert=op[PY_1BARS_LEFTVERT],
                                                 hor=op[PY_1BARS_LEFTHOR],
                                                 height2=op[PY_1BARS_LEFTMETHOD2DISTANCE],
                                                 )

            elif op[PY_1BARS_LEFT] == 2:
                # Dual Grid
                bars_container = self.bars_maker(op, width=op[PY_WINDOW_WIDTH],
                                                 height=op[PY_WINDOW_HEIGHT],
                                                 method=op[PY_1BARS_LEFT],
                                                 bank=op[PY_BANK],
                                                 vert=op[PY_1BARS_LEFTDUALVERT],
                                                 hor=op[PY_1BARS_LEFTDUALHOR],
                                                 height2=op[PY_1BARS_LEFTMETHOD2DISTANCE],
                                                 count=op[PY_1BARS_LEFTVERT])

            elif op[PY_1BARS_LEFT] == 3:
                # Star
                bars_container = self.bars_maker(op, width=op[PY_WINDOW_WIDTH], height=op[PY_WINDOW_HEIGHT],
                                                 method=op[PY_1BARS_LEFT],
                                                 bank=op[PY_BANK],
                                                 vert=op[PY_1BARS_LEFTSTARSTREAKS],
                                                 hor=op[PY_1BARS_LEFTSTARSUBD],
                                                 height2=op[PY_1BARS_LEFTSTARRADIUS],
                                                 count=op[PY_1BARS_LEFTSTARSTREAKS])

            self.bar_cutter(op, frames, bars_container)
            # bars_container = frames.GetDown().GetDown().GetDown().GetDown()
            # if bars_container.IsInstanceOf(c4d.Oboole):
            #     bars_container.Remove()


        # Bars End ==========================================================

        if op[PY_BANK]:
            # sill = self.make_sill(op)
            sill = self.create_sill(op)
            sill.InsertUnder(frames.GetDown())

        frames.InsertUnder(container)

        if op[PY_OFFSET] is not None:
            window = container.GetDown().GetDown()#.GetClone()
            window.SetRelPos(c4d.Vector(0, 0, op[PY_OFFSET]))

        return container

    @staticmethod
    def preshious_glass(op, obj, fixed=0, toplightarch=False):
        direction = 1.5
        normal = False
        if fixed != 4:
            if toplightarch:
                direction = -1.5
                normal = True
            else:
                direction = 1.5
                normal = False

            frame: c4d.BaseObject = obj.GetDown().GetDown().GetDown()
            numbers = frame.GetPointCount() // 21
            points_list = []
            start = 7
            for point in range(numbers):
                points_list.append(frame.GetPoint(start))
                start += 21
        else:
            # print(obj.GetDown())
            direction = 1.5
            normal = False
            frame: c4d.BaseObject = obj.GetDown()
            numbers = frame.GetPointCount() // 12
            points_list = []
            start = 4
            for point in range(numbers):
                points_list.append(frame.GetPoint(start))
                start += 12

        glass_spline = c4d.SplineObject(len(points_list), type=c4d.SPLINETYPE_LINEAR)

        glass_spline.SetAllPoints(points_list)
        glass_spline[c4d.SPLINEOBJECT_CLOSED] = True
        extrude = c4d.BaseObject(c4d.Oextrude)
        extrude[c4d.EXTRUDEOBJECT_FLIPNORMALS] = normal
        extrude[c4d.EXTRUDEOBJECT_EXTRUSIONOFFSET] = direction
        extrude.SetName("Glass")
        glass_spline.InsertUnder(extrude)
        extrude: c4d.PolygonObject = c4d.utils.SendModelingCommand(command=c4d.MCOMMAND_CURRENTSTATETOOBJECT,
                                                                   list=[extrude],
                                                                   doc=op.GetDocument())[0]

        for tag in extrude.GetTags():
            if tag.IsInstanceOf(c4d.Tpolygonselection) or tag.IsInstanceOf(c4d.Tedgeselection):
                tag.Remove()

        WindowHelper().catch_tex_tag(op, "glass", extrude)

        extrude.InsertUnder(frame)

    def desc_arch(self, op, single_id, description, flags):
        arch_height = c4d.DescID(PY_ARCH_HEIGHT)  # using ID from above post
        height = c4d.DescID(PY_WINDOW_HEIGHT)
        width = c4d.DescID(PY_WINDOW_WIDTH)

        # Begrenzung =======================================

        if single_id is None or width.IsPartOf(single_id)[0]:
            db = description.GetParameterI(width)

            width_limiter = PROFILE_WIDTH * 2

            db.SetFloat(c4d.DESC_MIN, width_limiter)

        if op[PY_ARCH_HEIGHT] is not None:
            if single_id is None or height.IsPartOf(single_id)[0]:
                db2 = description.GetParameterI(PY_WINDOW_HEIGHT)
                width_limiter = op[PY_WINDOW_WIDTH] / 1

                db2.SetFloat(c4d.DESC_MIN, op[PY_ARCH_HEIGHT] + width_limiter)

        # Hiding the Bars
        vert = c4d.DescID(PY_1BARS_LEFTVERT)
        hor = c4d.DescID(PY_1BARS_LEFTHOR)
        vert2 = c4d.DescID(PY_1BARS_LEFTDUALVERT)
        hor2 = c4d.DescID(PY_1BARS_LEFTDUALHOR)
        radius = c4d.DescID(PY_1BARS_LEFTSTARRADIUS)
        height = c4d.DescID(PY_1BARS_LEFTMETHOD2DISTANCE)
        streaks = c4d.DescID(PY_1BARS_LEFTSTARSTREAKS)
        segments = c4d.DescID(PY_1BARS_LEFTSTARSUBD)

        # left segment
        if single_id is None or vert.IsPartOf(single_id)[0]:
            vert_bc = description.GetParameterI(PY_1BARS_LEFTVERT)
            if op[PY_1BARS_LEFT] != 1:
                vert_bc.SetBool(c4d.DESC_HIDE, True)
        if single_id is None or hor.IsPartOf(single_id)[0]:
            hor_bc = description.GetParameterI(PY_1BARS_LEFTHOR)
            if op[PY_1BARS_LEFT] != 1:
                hor_bc.SetBool(c4d.DESC_HIDE, True)
        if single_id is None or vert2.IsPartOf(single_id)[0]:
            vert2_bc = description.GetParameterI(PY_1BARS_LEFTDUALVERT)
            if op[PY_1BARS_LEFT] != 2:
                vert2_bc.SetBool(c4d.DESC_HIDE, True)
        if single_id is None or hor2.IsPartOf(single_id)[0]:
            hor2_bc = description.GetParameterI(PY_1BARS_LEFTDUALHOR)
            if op[PY_1BARS_LEFT] != 2:
                hor2_bc.SetBool(c4d.DESC_HIDE, True)
        if single_id is None or streaks.IsPartOf(single_id)[0]:
            streaks_bc = description.GetParameterI(PY_1BARS_LEFTSTARSTREAKS)
            if op[PY_1BARS_LEFT] != 3:
                streaks_bc.SetBool(c4d.DESC_HIDE, True)
        if single_id is None or segments.IsPartOf(single_id)[0]:
            segments_bc = description.GetParameterI(PY_1BARS_LEFTSTARSUBD)
            if op[PY_1BARS_LEFT] != 3:
                segments_bc.SetBool(c4d.DESC_HIDE, True)
        if single_id is None or radius.IsPartOf(single_id)[0]:
            radius_bc = description.GetParameterI(PY_1BARS_LEFTSTARRADIUS)
            if op[PY_1BARS_LEFT] != 3:
                radius_bc.SetBool(c4d.DESC_HIDE, True)
        if single_id is None or height.IsPartOf(single_id)[0]:
            height_bc = description.GetParameterI(PY_1BARS_LEFTMETHOD2DISTANCE)
            if op[PY_1BARS_LEFT] == 0 or op[PY_1BARS_LEFT] == 3:
                height_bc.SetBool(c4d.DESC_HIDE, True)

class Windows2023(plugins.ObjectData, WindowHelper):

    def __init__(self):
        self.SetOptimizeCache(True)
        self.profile = self.load_profile()
        self.handle = self.load_handle()
        self.sill = self.load_sill()

    def Init(self, op):

        # Global Paramters
        self.InitAttr(op, bool, PY_BANK)
        self.InitAttr(op, float, PY_BANK_SIZE)
        self.InitAttr(op, float, PY_BANK_INNERSIZE)
        self.InitAttr(op, float, PY_BAR_WIDTH)

        # Arch Window Paramters

        self.InitAttr(op, float, PY_ARCH_HEIGHT)
        self.InitAttr(op, int, PY_ARCH_SUBD)
        self.InitAttr(op, bool, PY_TILT_CHECK)
        self.InitAttr(op, float, PY_WINDOW_OPEN)
        self.InitAttr(op, bool, PY_WINDOW_DIRECTION)
        self.InitAttr(op, bool, PY_ARCH_POINTED)
        self.InitAttr(op, bool, PY_ARCH_FIXER)
        self.InitAttr(op, float, PY_ARCH_STRENGTH)

        # Bars Arch left
        self.InitAttr(op, int, PY_1BARS_LEFT)
        self.InitAttr(op, int, PY_1BARS_LEFTVERT)
        self.InitAttr(op, int, PY_1BARS_LEFTHOR)
        self.InitAttr(op, float, PY_1BARS_LEFTMETHOD2DISTANCE)
        self.InitAttr(op, float, PY_1BARS_LEFTSTARRADIUS)
        self.InitAttr(op, int, PY_1BARS_LEFTDUALVERT)
        self.InitAttr(op, int, PY_1BARS_LEFTDUALHOR)
        self.InitAttr(op, int, PY_1BARS_LEFTSTARSTREAKS)
        self.InitAttr(op, int, PY_1BARS_LEFTSTARSUBD)
        
        # Global Parameters
        op[PY_WINDOW_LIST] = 1
        op[PY_WINDOW_WIDTH] = 100
        op[PY_WINDOW_HEIGHT] = 120

        op[PY_BANK] = False
        op[PY_BANK_SIZE] = 0.0
        op[PY_BANK_INNERSIZE] = 0.0
        op[PY_BAR_WIDTH] = 2.0

        # Arch Window
        op[PY_WINDOW_SEGMENTS] = False
        op[PY_ARCH_POINTED] = False
        op[PY_ARCH_FIXER] = False
        op[PY_ARCH_STRENGTH] = 30
        op[PY_ARCH_HEIGHT] = 20
        op[PY_ARCH_SUBD] = 8
        op[PY_TILT_CHECK] = False
        op[PY_WINDOW_OPEN] = 0.0
        op[PY_WINDOW_DIRECTION] = False

        # Bars Arch Window left
        op[PY_1BARS_LEFT] = 0
        op[PY_1BARS_LEFTVERT] = 1
        op[PY_1BARS_LEFTHOR] = 1
        op[PY_1BARS_LEFTMETHOD2DISTANCE] = 50
        op[PY_1BARS_LEFTSTARRADIUS] = 20
        op[PY_1BARS_LEFTDUALVERT] = 3
        op[PY_1BARS_LEFTDUALHOR] = 1
        op[PY_1BARS_LEFTSTARSTREAKS] = 4
        op[PY_1BARS_LEFTSTARSUBD] = 2

        # op[c4d.MSG_CHANGE]
        return True

    def Message(self, node, type, data):

        if type == c4d.MSG_MENUPREPARE:
            self.create_tex_tags(node, "Sill")
            self.create_tex_tags(node, "Handle")
            self.create_tex_tags(node, "Bars")
            self.create_tex_tags(node, "Glass")
            self.create_tex_tags(node, "Frame")

        return True

    def GetVirtualObjects(self, op, hh):
        # dirty = True if a cache is dirty or if the data (any parameters) of the object changed.
        # If nothing changed and a cache is present, return the cache

        # cache deaktiveren

        # dirty = op.CheckCache(hh) or op.IsDirty(c4d.DIRTYFLAGS_DESCRIPTION)
        # if not dirty:
        #     return op.GetCache(hh)

        """Settings for Arch Window"""

        return self.window_arch(op)

    def GetDDescription(self, op, description, flags):

        if not description.LoadDescription(op.GetType()):
            return False

        single_id = description.GetSingleDescID()

        # Window 1 : Arch Window ==========================
        if op[PY_WINDOW_LIST] == 1:
            self.desc_arch(op, single_id, description, flags)

        return True, flags | c4d.DESCFLAGS_DESC_LOADED

    def GetDEnabling(self, op, did, t_data, flags, itemdesc):

        # Window 1: Arch Window
        if op[PY_WINDOW_LIST] == 1:
            if did[0].id == PY_1BARS_LEFTMETHOD2DISTANCE:
                if op[PY_1BARS_LEFT] <= 1:
                    return False

        return True


if __name__ == "__main__":

    path, file = os.path.split(__file__)
    files = "icon.tif"
    new_path = os.path.join(path, "res", files)
    bitmap = bitmaps.BaseBitmap()
    bitmap.InitWith(new_path)

    icons_path = os.path.join(path, "res", "iconslist.tif")
    icons = bitmaps.BaseBitmap()
    icons.InitWith(icons_path)
    c4d.gui.RegisterIcon(ICON_ID, icons, x=0, y=0, w=32, h=32)
    c4d.gui.RegisterIcon(ICON_ID_STOP, icons, x=128, y=0, w=32, h=32)
    c4d.gui.RegisterIcon(ICON_ID_GRID, icons, x=224, y=0, w=32, h=32)
    c4d.gui.RegisterIcon(ICON_ID_DUALGRID, icons, x=256, y=0, w=32, h=32)
    c4d.gui.RegisterIcon(ICON_ID_STAR, icons, x=288, y=0, w=32, h=32)
    c4d.gui.RegisterIcon(ICON_ID_BALCONY, icons, x=0, y=32, w=32, h=32)

    if plugins.RegisterObjectPlugin(id=PLUGIN_ID,
                                    str="C4D Windows 2023",
                                    g=Windows2023,
                                    description="windows2023",
                                    icon=bitmap,
                                    info=c4d.OBJECT_GENERATOR,
                                    disklevel=1):
        print("C4D Windows 2023.2 -----> started")
    else:

        print("C4D Windows 2023.2 register failed")
